/* -*-c-*-
 *
 * Fake rmt(8) server for hashing and storing files
 *
 * (c) 2010 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of the distorted.org.uk backup suite.
 *
 * distorted-backup is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * distorted-backup is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with distorted-backup; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#define _FILE_OFFSET_BITS 64

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#include <getopt.h>

#include <pwd.h>

#include <mLib/dstr.h>
#include <mLib/quis.h>
#include <mLib/report.h>

#include <nettle/sha.h>

/*----- Configuration -----------------------------------------------------*/

#ifndef BKP
#  define BKP "/mnt/bkp"
#endif

/*----- Main code ---------------------------------------------------------*/

#define BUFSZ 10240
#define LSEEK_GET_TAPEPOS 10
#define LSEEK_GO2_TAPEPOS 11

struct flag {
  const char *name;
  int f;
};

static const struct flag openflag[] = {
  /* ;;; Emacs Lisp to generate the table below.  Place your cursor just
     ;;; after the closing `)' and press C-x C-e.

     (let ((flags '(rdonly wronly rdwr creat excl trunc nonblock ndelay
		    noctty append dsync rsync sync cloexec async
		    direct noatime nofollow shlock exlock defer)))
       (save-excursion
	 (goto-char (point-min))
	 (search-forward (concat "***" "BEGIN openflag" "***"))
	 (beginning-of-line 2)
	 (delete-region (point)
			(progn
			  (search-forward "***END***")
			  (beginning-of-line)
			  (point)))
       (dolist (f (sort (copy-list flags) #'string<))
	 (let ((up (upcase (symbol-name f))))
	   (insert (format "#ifdef O_%s\n" up))
	   (insert (format "  { \"%s\", O_%s },\n" up up))
	   (insert "#endif\n")))))
  */
  /***BEGIN openflag***/
#ifdef O_APPEND
  { "APPEND", O_APPEND },
#endif
#ifdef O_ASYNC
  { "ASYNC", O_ASYNC },
#endif
#ifdef O_CLOEXEC
  { "CLOEXEC", O_CLOEXEC },
#endif
#ifdef O_CREAT
  { "CREAT", O_CREAT },
#endif
#ifdef O_DEFER
  { "DEFER", O_DEFER },
#endif
#ifdef O_DIRECT
  { "DIRECT", O_DIRECT },
#endif
#ifdef O_DSYNC
  { "DSYNC", O_DSYNC },
#endif
#ifdef O_EXCL
  { "EXCL", O_EXCL },
#endif
#ifdef O_EXLOCK
  { "EXLOCK", O_EXLOCK },
#endif
#ifdef O_NDELAY
  { "NDELAY", O_NDELAY },
#endif
#ifdef O_NOATIME
  { "NOATIME", O_NOATIME },
#endif
#ifdef O_NOCTTY
  { "NOCTTY", O_NOCTTY },
#endif
#ifdef O_NOFOLLOW
  { "NOFOLLOW", O_NOFOLLOW },
#endif
#ifdef O_NONBLOCK
  { "NONBLOCK", O_NONBLOCK },
#endif
#ifdef O_RDONLY
  { "RDONLY", O_RDONLY },
#endif
#ifdef O_RDWR
  { "RDWR", O_RDWR },
#endif
#ifdef O_RSYNC
  { "RSYNC", O_RSYNC },
#endif
#ifdef O_SHLOCK
  { "SHLOCK", O_SHLOCK },
#endif
#ifdef O_SYNC
  { "SYNC", O_SYNC },
#endif
#ifdef O_TRUNC
  { "TRUNC", O_TRUNC },
#endif
#ifdef O_WRONLY
  { "WRONLY", O_WRONLY },
#endif
  /***END***/
  { 0, 0 }
};

int main(int argc, char *argv[])
{
  int ch;
  dstr d = DSTR_INIT, dd = DSTR_INIT;
  unsigned char buf[BUFSZ];
  int fd = -1, hfd = -1;
  off_t rc;
  off_t off = 0;
  struct passwd *pw;
  uid_t u;
  unsigned f = 0;
#define f_bogus 1u
  const char *p = 0;
  const char *bkp = 0, *host = 0;
  struct sha256_ctx hc;

  ego(argv[0]);
  setvbuf(stdin, 0, _IONBF, 0);
  signal(SIGPIPE, SIG_IGN);

  for (;;) {
    int o = getopt(argc, argv, "H:r:");
    if (o < 0) break;
    switch (o) {
      case 'H': host = optarg; break;
      case 'r': bkp = optarg; break;
      default: f |= f_bogus; break;
    }
  }
  argc -= optind; argv += optind;
  if ((f & f_bogus) || argc) {
    pquis(stderr, "usage: $ [-r ROOT] [-H HOST]\n");
    exit(1);
  }

  if (!bkp) bkp = getenv("BKP");
  if (!bkp) bkp = BKP;

  if (!host) host = getenv("BKP_HOST");

  if (!host) {
    p = getenv("USER");
    if (!p) p = getenv("LOGNAME");
    if (!p) {
      u = getuid();
      if ((pw = getpwuid(u)) == 0)
	die(1, "no passwd entry (you don't exist?)");
      p = pw->pw_name;
    }
    if (strncmp(p, "bkp-", 4) != 0) {
      die(1, "can't deduce host name: "
	  "login name `%s' doesn't begin with `bkp-'",
	  p);
    }
    host = p + 4;
  }

  for (;;) {
    if (fflush(stdout)) goto fail;
    ch = getchar();
    if (ch == EOF) break;
    DRESET(&d); DRESET(&dd); rc = 0;

#define CHECKFD do { if (fd < 0) goto badf; } while (0)
#define ERROR(what) do moan(what ": %s", strerror(errno)); while (0)
#define ERROR1(what, arg) \
  do moan(what ": %s", arg, strerror(errno)); while (0)
#define ARG(d) do {							\
  if (dstr_putline(&d, stdin) == EOF || ferror(stdin) || feof(stdin))	\
    { moan("read (stdin)", strerror(errno)); goto fail; }		\
} while (0)

#define SKIPWS do { while (isspace((unsigned char)*p)) p++; } while (0)

    switch (ch) {

      case 'O': {
	/* Ofile\nmode\n -- open file */

	const struct flag *ff;
	char *p, *q;
	size_t n;
	long mode, f;

	ARG(d); ARG(dd);
	if (fd >= 0 && close(fd)) ERROR("close (fd)");
	if (hfd >= 0 && close(hfd)) ERROR("close (hash)");
	fd = hfd = -1;

	if (chdir(bkp) || chdir(host)) ERROR1("chdir (%s)", host);
	p = d.buf;
	if ((q = strchr(p, '/')) == 0)
	  { moan("bad path: missing `/')"); goto inval; }
	*q++ = 0;
	if (chdir(p) || chdir("prepare/incoming")) ERROR1("chdir (%s)", p);

	memmove(d.buf, q, d.len - (q - d.buf) + 1);
	d.len -= q - d.buf;

	errno = 0;
	mode = strtol(dd.buf, &p, 0);
	if (errno) ERROR("bad mode");
	else if (mode < 0 || mode > INT_MAX)
	  { moan("bad mode: range"); goto range; }
	SKIPWS;
	if (!*p) {
	  switch (mode & O_ACCMODE) {
	    case O_RDONLY: mode = O_RDONLY; break;
	    case O_WRONLY: mode = O_WRONLY | O_TRUNC | O_CREAT; break;
	    case O_RDWR: mode = O_RDWR;
	    default: moan("bad mode: unknown access type"); goto inval;
	  }
	} else {
	  mode = 0;
	  for (;;) {
	    if (p[0] == 'O' && p[1] == '_') {
	      p += 2;
	      n = strcspn(p, " \t|");
	      for (ff = openflag; ff->name; ff++) {
		if (strncmp(p, ff->name, n) == 0 && !ff->name[n])
		  goto ofmatch;
	      }
	      moan("bad mode: unknown flag O_%.*s", (int)n, p);
	      goto inval;
	    ofmatch:
	      mode |= ff->f;
	      p += n;
	    } else if (isdigit((unsigned long)*p)) {
	      errno = 0;
	      f = strtol(p, &p, 0);
	      if (errno) ERROR("bad mode");
	      else if (f < 0 || f > INT_MAX)
		{ moan("bad mode: range"); goto range; }
	      mode |= f;
	    } else {
	      moan("bad mode: unexpected token");
	      goto inval;
	    }
	    SKIPWS;
	    if (!*p)
	      break;
	    else if (*p != '|') {
	      moan("bad mode: expected `|'");
	      goto inval;
	    }
	    p++;
	    SKIPWS;
	  }
	}

	if ((fd = open(d.buf, mode, 0666)) < 0) ERROR1("open (%s)", d.buf);
	if ((mode & O_ACCMODE) == O_WRONLY) {
	  DPUTS(&d, ".hash");
	  if ((hfd = open(d.buf,
			 mode & (O_ACCMODE | O_CREAT | O_EXCL | O_TRUNC),
			 0666)) < 0) {
	    close(fd); fd = -1;
	    ERROR1("open (%s)", d.buf);
	  }
	  sha256_init(&hc);
	}
	off = 0;
      } break;

      case 'C': {
	/* Chunoz\n -- close file */

	uint8_t h[SHA256_DIGEST_SIZE], *p;
	char hex[SHA256_DIGEST_SIZE * 2 + 1], *q;
	unsigned i;

	ARG(d); CHECKFD;
	if (close(fd)) ERROR("close (fd)");
	fd = -1;
	if (hfd >= 0) {
	  sha256_digest(&hc, sizeof(h), h);
	  for (p = h, q = hex; p < h + sizeof(h); p++, q += 2)
	    sprintf(q, "%02x", *p);
	  *q++ = '\n';
	  errno = EIO;
	  if (write(hfd, hex, sizeof(hex)) < sizeof(hex) || close(hfd))
	    ERROR("close (hash)");
	  hfd = -1;
	}
      } break;

      case 'L': {
	/* Loffset\nwhence\n -- seek
	 *  (warning: the manual page gets these the wrong way round)
	 */

	int whence;
	off_t offset;

	ARG(d); ARG(dd); CHECKFD;
	offset = atoi(d.buf); whence = strtoull(dd.buf, 0, 0);
	switch (whence) {
	  case LSEEK_GET_TAPEPOS: whence = SEEK_CUR; offset = 0; break;
	  case LSEEK_GO2_TAPEPOS: whence = SEEK_SET; break;
	}
	switch (whence) {
	  case SEEK_CUR:
	    if (!offset) { rc = off; break; }
	  default:
	    rc = lseek(fd, offset, whence);
	    if (rc == (off_t)-1) ERROR("seek");
	    off = rc;
	    break;
	}
      } break;

      case 'W': {
	/* Wlen\ndata... -- write */

	size_t n, nn, sz;
	ssize_t ssz;
	unsigned char *p;
	int botch = 0;

	ARG(d); CHECKFD;
	rc = sz = strtoul(d.buf, 0, 0);
	while (sz) {
	  nn = sz > BUFSZ ? BUFSZ : sz;
	  n = fread(buf, 1, nn, stdin);
	  if (n < nn) {
	    if (feof(stdin)) { moan("eof on stdin"); goto fail;}
	    else ERROR("read (stdin)");
	  }
	  if (hfd >= 0) sha256_update(&hc, n, buf);
	  p = buf;
	  while (!botch && n) {
	    if ((ssz = write(fd, p, n)) > 0) {
	      p += ssz; off += ssz; n -= ssz;
	    } else if (!ssz) { moan("zero-length write"); goto fail; }
	    else if (errno != EINTR) { botch = errno; }
	  }
	  sz -= nn;
	}
	if (botch) { errno = botch; ERROR("write"); }
      } break;

      case 'R': {
	/* Rlen\n -- read */

	size_t nn;
	ssize_t ssz;

	ARG(d); CHECKFD;
	nn = strtoul(d.buf, 0, 0); if (nn > BUFSZ) nn = BUFSZ;
	if ((ssz = read(fd, buf, nn)) < 0) ERROR("read");
	off += ssz;
	printf("A%ld\n", (long)ssz);
	if (fwrite(buf, 1, ssz, stdout) < ssz)
	  { moan("write (stdout): %s", strerror(errno)); goto fail; }
	continue;
      } break;

      case 'i': case 'I':
	/* Iop\ncount\n -- ioctl */
	ARG(d); ARG(dd); CHECKFD; goto notty;

      case 'S':
	/* S -- ioctl */
	CHECKFD; goto notty;
      case 's':
	/* sop -- ioctl */

	if ((ch = getchar()) == EOF) goto fail;
	CHECKFD; goto notty;

      default:
	goto fail;
    }

    printf("A%llu\n", (unsigned long long)rc);
    continue;

  badf: errno = EBADF; goto error;
  range: errno = ERANGE; goto error;
  inval: errno = EINVAL; goto error;
  notty: errno = ENOTTY; goto error;
  error:
    printf("E%d\n%s\n", errno, strerror(errno));
    continue;
  }
  if (fflush(stdout) || ferror(stdout) || ferror(stdin)) goto fail;
  return (0);
fail:
  return (1);
}

/*----- That's all, folks -------------------------------------------------*/
